__all__ = [
    'FileOperation',
]

import pythoncom
from win32com.shell import shell  # type: ignore (module *does* exist)

from .interfaces import IFileOperation


def FileOperation() -> IFileOperation:
    return pythoncom.CoCreateInstance(
        shell.CLSID_FileOperation, None, pythoncom.CLSCTX_ALL, shell.IID_IFileOperation
    )  # type: ignore



__all__ = [
    'IFileOperationError',
    'FileOperatorError',
    'UserCancelledError',
    'InterfaceNotImplementedError',
    # Backwards compatible exceptions:
    'IFO_NotADirectoryError',
]

import struct


def int32_to_uint32(value: int) -> int:
    """pythoncom defines its HRESULTs as signed integers for some reason."""
    return struct.unpack('I', struct.pack('i', value))[0]


class IFileOperationError(Exception):
    pass


class InterfaceNotImplementedError(IFileOperationError):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            'COM interface method not implemented.  This is likely a WINE bug, see: '
            'https://bugs.winehq.org/show_bug.cgi?id=50064.'
        )


class FileOperatorError(IFileOperationError):
    def __init__(
        self,
        hresult: int = 0,
        msg: str = '',
        excepinfo=None,
        argerror=None,
        *args,
        **kwargs,
    ) -> None:
        if hresult < 0:
            hresult = int32_to_uint32(hresult)
        self.hresult = hresult
        if excepinfo or argerror:
            msg = f'{hex(hresult)}: {msg}, {excepinfo}, {argerror}'
        else:
            msg = f'{hex(hresult)}: {msg}'
        super().__init__(msg)


class UserCancelledError(IFileOperationError):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__('User cancelled the operation')


class UnexpectedError(IFileOperationError):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(f'An unexpected error occurred: {args}, {kwargs}')


# Backward Campatible exceptions: remove these in the next major version (2.0)
# These are for backwards campatibility when replacing specific HRESULT
# based exceptions with standard library exceptions
class IFO_NotADirectoryError(FileOperatorError, NotADirectoryError):
    """FileOperatorError with HRESULT E_DRIVE_NOT_FOUND"""



__all__ = [
    'FileAttributeFlags',
    'FileOperationFlags',
    'TransferSourceFlags',
]

from enum import Flag, IntEnum, IntFlag


class FileAttributeFlags(IntFlag):
    """Flags to use on a WIN32_FIND_DATAW structured to set the file attributes. See:
    https://learn.microsoft.com/en-us/windows/win32/fileio/file-attribute-constants"""

    ARCHIVE = 0x20
    """A file or directory that is an archive file or directory. Applications typically
    use this attribute to mark files for backup or removal."""

    COMPRESSED = 0x800
    """A file or directory that is compressed. For a file, all of the data in the file
    is compressed. For a directory, compression is the default for newly created files
    and subdirectories."""

    DEVICE = 0x40
    """This value is reserved for system use."""

    DIRECTORY = 0x10
    """The handle that identifies a directory."""

    ENCRYPTED = 0x4000
    """A file or directory that is encrypted. For a file, all data streams in the file
    are encrypted. For a directory, encryption is the default for newly created files
    and subdirectories."""

    HIDDEN = 0x2
    """The file or directory is hidden. It is not included in an ordinary directory
    listing."""

    INTEGRITY_STREAM = 0x8000
    """The directory or user data stream is configured with integrity (only supported on
    ReFS volumes). It is not included in an ordinary directory listing. The integrity
    setting persists with the file if it's renamed. If a file is copied the destination
    file will have integrity set if either the source file or destination directory have
    integrity set."""

    NORMAL = 0x80
    """A file that does not have other attributes set. This attribute is valid only when
    used alone."""

    NOT_CONTENT_INDEXED = 0x2000
    """The file or directory is not to be indexed by the content indexing service."""

    NO_SCRUB_DATA = 0x20000
    """The user data stream not to be read by the background data integrity scanner (AKA
    scrubber). When set on a directory it only provides inheritance. This flag is only
    supported on Storage Spaces and ReFS volumes. It is not included in an ordinary
    directory listing."""

    OFFLINE = 0x1000
    """The data of a file is not available immediately. This attribute indicates that
    the file data is physically moved to offline storage. This attribute is used by
    Remote Storage, which is the hierarchical storage management software. Applications
    should not arbitrarily change this attribute."""

    READONLY = 0x1
    """A file that is read-only. Applications can read the file, but cannot write to it
    or delete it. This attribute is not honored on directories. """

    RECALL_ON_DATA_ACCESS = 0x400000
    """When this attribute is set, it means that the file or directory is not fully
    present locally. For a file that means that not all of its data is on local storage
    (e.g. it may be sparse with some data still in remote storage). For a directory it
    means that some of the directory contents are being virtualized from another
    location. Reading the file / enumerating the directory will be more expensive than
    normal, e.g. it will cause at least some of the file/directory content to be fetched
    from a remote store. Only kernel-mode callers can set this bit."""

    RECALL_ON_OPEN = 0x40000
    """This attribute only appears in directory enumeration classes
    (FILE_DIRECTORY_INFORMATION, FILE_BOTH_DIR_INFORMATION, etc.). When this attribute
    is set, it means that the file or directory has no physical representation on the
    local system; the item is virtual. Opening the item will be more expensive than
    normal, e.g. it will cause at least some of it to be fetched from a remote store."""

    REPARSE_POINT = 0x400
    """A file or directory that has an associated reparse point, or a file that is a
    symbolic link."""

    SPARSE_FILE = 0x200
    """A file that is a sparse file."""

    SYSTEM = 0x4
    """A file or directory that the operating system uses a part of, or uses
    exclusively."""

    TEMPORARY = 0x100
    """A file that is being used for temporary storage. File systems avoid writing data
    back to mass storage if sufficient cache memory is available, because typically, an
    application deletes a temporary file after the handle is closed. In that scenario,
    the system can entirely avoid writing the data. Otherwise, the data is written after
    the handle is closed."""

    VIRTUAL = 0x10000
    """This value is reserved for system use."""

    PINNED = 0x80000
    """This attribute indicates user intent that the file or directory should be kept
    fully present locally even when not being actively accessed. This attribute is for
    use with hierarchical storage management software."""

    UNPINNED = 0x200000
    """This attribute indicates that the file or directory should not be kept fully
    present locally except when being actively accessed. This attribute is for use with
    hierarchical storage management software."""


class FileOperationFlags(IntFlag):
    ALLOW_UNDO = 0x40
    """Allow the operation to be undone."""

    FILES_ONLY = 0x80
    """Only operate on files, not folders, if a wildcard is used."""

    NO_CONFIMATION = 0x10
    """Automatically respond to any dialog box that would be shown with 'Yes to All'."""

    NO_CONFIRM_MKDIR = 0x200
    """Do not confirm the creation of a new directory if the operation requires one to
    be created."""

    NO_CONNECTED_ELEMENTS = 0x2000
    """Do not move connected elements as a group."""

    NOCOPYSECURITYATTRIBS = 0x800
    """Do not copy the security attributes of the file."""

    NOERRORUI = 0x400
    """Do not display a message to the user if an error occurs. If this flag is set
    without EARLYFAILURE, any error is treated as if the user had chosen 'Ignore' or
    'Continue'. It halts the current action, sets a flag to indicate that an action was
    aborted, and proceeds with the rest of the operation."""

    NORECURSION = 0x8000
    """Only operate in the local folder. Do not operate recursively into subdirectories.
    """

    RENAMEONCOLLISION = 0x8
    """Give the item being operated on a new name in a move, copy, or rename operation
    if an item with the target name already exists."""

    SILENT = 0x4
    """Do not display a progress dialog box."""

    WANTNUKEWARNING = 0x1000
    """Send a warning if a file or folder is being destroyed during a delete operation
    rather than recycled. This flag partially overrides NOCONFIRMATION."""

    ADDUNDORECORD = 0x20000000
    """Introduced in Windows 8. The file operation was user-invoked and should be placed
    on the undo stack. This flag is preferred to ALLOWUNDO."""

    NOSKIPJUNCTIONS = 0x10000
    """Walk into Shell namespace junctions. By default, junctions are not entered."""

    PREFERHARDLINK = 0x20000
    """If possible, create a hard link rather than a new instance of the file in the
    destination."""

    SHOWELEVATIONPROMPT = 0x40000
    """If an operation requires elevated rights and the NOERRORUI flag is set to disable
    error UI, display a UAC UI prompt nonetheless."""

    EARLYFAILURE = 0x100000
    """If EARLYFAILURE and NOERRORUI are set, the entire set of operations is stopped
    upon encountering any error in any operation."""

    PRESERVEFILEEXTENSIONS = 0x200000
    """With RENAMEONCOLLISION, rename collisions in such a way as to preserve file name
    extensions."""

    KEEPNEWERFILE = 0x400000
    """Keep the newer file or folder, based on the Date Modified property, if a
    collision occurs. This is done automatically with no prompt UI presented to the
    user."""

    NOCOPYHOOKS = 0x800000
    """Do not use copy hooks."""

    NOMINIMIZEBOX = 0x1000000
    """Do not allow the progress dialog to be minimized."""

    MOVEACLSACROSSVOLUMES = 0x2000000
    """Copy the security attributes of the source item to the destination item when
    performing a cross-volume move operation. Without this flag, the destination item
    receives the security attributes of its new folder."""

    DONTDISPLAYSOURCEPATH = 0x4000000
    """Do not display the path of the source item in the progress dialog."""

    DONTDISPLAYDESTPATH = 0x8000000
    """Do not display the path of the destination item in the progress dialog."""

    RECYCLEONDELETE = 0x80000
    """Introduced in Windows 8. When a file is deleted, send it to the Recycle Bin
    rather than permanently deleting it."""

    REQUIREELEVATION = 0x10000000
    """Introduced in Windows Vista SP1. The user expects a requirement for rights
    elevation, so do not display a dialog box asking for a confirmation of the
    elevation."""

    COPYASDOWNLOAD = 0x40000000
    """Introduced in Windows 7. Display a Downloading instead of Copying message in the
    progress dialog."""

    DONTDISPLAYLOCATIONS = 0x80000000
    """Introduced in Windows 7. Do not display the location line in the progress dialog.
    """


class TransferSourceFlags(Flag):
    """Flags recieved in a Pre/Post-Move/Copy, etc event in IFileOperationProgressSink.
    https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/ne-shobjidl_core-_transfer_source_flags
    """

    # Just a Flag, not IntFlag, since we don't pass it *into* any win32 API calls.

    NORMAL = 0x0
    FAIL_EXIST = 0x0
    """Fail if the destination already exists, unless OVERWRITE_EXIST is specified. This
    is a default behavior."""

    RENAME_EXIST = 0x1
    """Rename with auto-name generation if the destination already exists."""

    OVERWRITE_EXIST = 0x2
    """Overwrite or merge with the destination."""

    ALLOW_DECRYPTION = 0x4
    """Allow creation of a decrypted destination."""

    NO_SECURITY = 0x8
    """No discretionary access control list (DACL), system access control list (SACL),
    or owner."""

    COPY_CREATION_TIME = 0x10
    """Copy the creation time as part of the copy. This can be useful for a move
    operation that is being used as a copy and delete operation
    (MOVE_AS_COPY_DELETE)."""

    COPY_WRITE_TIME = 0x20
    """Copy the last write time as part of the copy."""

    USE_FULL_ACCESS = 0x40
    """Assign write, read, and delete permissions as share mode."""

    DELETE_RECYCLE_IF_POSSIBLE = 0x80
    """Recycle on file delete, if possible."""

    COPY_HARD_LINK = 0x100
    """Hard link to the desired source (not required). This avoids a normal copy
    operation."""

    COPY_LOCALIZED_NAME = 0x200
    """Copy the localized name."""

    MOVE_AS_COPY_DELETE = 0x400
    """Move as a copy and delete operation."""

    TSF_SUSPEND_SHELLEVENTS = 0x800
    """Suspend Shell events."""


class FileOperationResult(IntEnum):
    # NOTE: I haven't been able to find this in the MSDocs anwhere. The closest I've
    # found is in pywin32 this is defined as COPYENGINE_S_DONT_PROCESS_CHILDREN, which
    # doesn't make sense as a return code from the operations it's coming from. But,
    # it's the result I'm getting for successful file operations.
    SUCCESS = 0x270008
    """A successful move/copy/rename/delete operation."""

    E_DESTINATION_IS_FILE = 0x8027000B
    """A move/copy/rename tried to operate on a folder, but the target already exists as
    a file.
    """

    E_DESTINATION_IS_FOLDER = 0x8027000C
    """A move/copy/rename tried to operate on a file, but the target already exists as a
    folder.
    """

    E_REQUIRES_ELEVATION = 0x80270002
    E_ACCESS_DENIED_SOURCE = 0x80270021
    E_ACCESS_DENIED_DESTINATION = 0x80270022
    """The user requires elevated permissions to perform the operation."""

    E_ALREADY_EXISTS_NORMAL = 0x80270029
    E_ALREADY_EXISTS_READONLY = 0x8027002A
    E_ALREADY_EXISTS_SYSTEM = 0x8027002B
    E_ALREADY_EXISTS_FOLDER = 0x8027002C
    """The file already exists."""

    E_USER_CANCELLED = 0x80270000
    """User cancelled the operation."""

    E_DRIVE_NOT_FOUND = 0x8027000F  # COPYENGINE_E_DEST_IS_RO_CD
    """The destination is a read-only CD, or the drive was not found."""

    E_NOT_IMPLEMENTED = 0x80004001
    """The method is not implemented."""

    E_UNEXPECTED = 0x8000FFFF
    """An unexpected error occurred."""



__all__ = [
    'IFileSysBindData',
]

from ctypes import POINTER
from ctypes.wintypes import PWIN32_FIND_DATAW

from comtypes import COMMETHOD, GUID, HRESULT, IUnknown


class IFileSysBindData(IUnknown):
    """Define the IFileSysBindData interface using comtypes, see
    https://learn.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-ifilesystembinddata
    https://github.com/tpn/winsdk-10/blob/9b69fd26ac0c7d0b83d378dba01080e93349c2ed/Include/10.0.16299.0/um/ShObjIdl_core.h#L24285
    """

    _iid_ = GUID('{01e18d10-4d8b-11d2-855d-006008059367}')
    _methods_ = [
        COMMETHOD([], HRESULT, 'SetFindData', (['in'], PWIN32_FIND_DATAW, 'pfd')),
        COMMETHOD(
            [], HRESULT, 'GetFindData', (['out'], POINTER(PWIN32_FIND_DATAW), 'ppfd')
        ),
    ]



"""
Python implementation/version of the IFileOperationProgressSink interface.
"""
from __future__ import annotations

__all__ = [
    'FileOperationProgressSink',
]

from functools import cached_property, wraps
from typing import Callable, ParamSpec, overload

from comtypes import COMObject
from comtypes.hresult import E_FAIL, S_OK

from ..flags import FileAttributeFlags, TransferSourceFlags
from .common import com_ptr
from .interfaces import IFileOperationProgressSink, IShellItem

P = ParamSpec('P')


@overload
def get_name(item: None) -> None:
    ...


@overload
def get_name(item: IShellItem) -> str:
    ...


def get_name(item: IShellItem | None) -> str | None:
    """Helper to get the filename from an IShellItem"""
    if not item:
        return None
    name = item.GetDisplayName(0x80058000)  # SIGDN_FILESYSPATH
    if not name:
        name = item.GetDisplayName(0)  # SIGDN_NORMALDISPLAY
    return name


def _err_to_hresult(method: Callable[P, None]) -> Callable[P, int]:
    @wraps(method)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> int:
        try:
            method(*args, **kwargs)
        except Exception:
            return E_FAIL
        return S_OK

    return wrapped


class _FileOperationProgressSink(COMObject):
    """Implementation of IFileOperationProgressSink, passes along calls to the
    Python implementation class FileOperationProgressSink.
    """

    _com_interfaces_ = [IFileOperationProgressSink]

    def __init__(self, impl: FileOperationProgressSink):
        self.impl = impl

    @_err_to_hresult
    def StartOperations(self):
        self.impl.start_operations()

    @_err_to_hresult
    def FinishOperations(self, hresult: int):
        self.impl.finish_operations(hresult)

    @_err_to_hresult
    def PreRenameItem(self, dwFlags: int, psiItem: IShellItem, pszNewName: str):
        self.impl.pre_rename_item(
            TransferSourceFlags(dwFlags), get_name(psiItem), pszNewName
        )

    @_err_to_hresult
    def PostRenameItem(
        self,
        dwFlags: int,
        psiItem: IShellItem,
        pszNewName: str,
        hrRename: int,
        psiNewlyCreated: IShellItem | None,
    ):
        self.impl.post_rename_item(
            TransferSourceFlags(dwFlags),
            get_name(psiItem),
            get_name(psiNewlyCreated),
            hrRename,
        )

    @_err_to_hresult
    def PreMoveItem(
        self,
        dwFlags: int,
        psiItem: IShellItem,
        psiDestinationFolder: IShellItem,
        pszNewName: str,
    ):
        self.impl.pre_move_item(
            TransferSourceFlags(dwFlags),
            get_name(psiItem),
            get_name(psiDestinationFolder),
            pszNewName,
        )

    @_err_to_hresult
    def PostMoveItem(
        self,
        dwFlags: int,
        psiItem: IShellItem,
        psiDestinationFolder: IShellItem,
        pszNewName: str,
        hrMove: int,
        psiNewlyCreated: IShellItem | None,
    ):
        self.impl.post_move_item(
            TransferSourceFlags(dwFlags),
            get_name(psiItem),
            get_name(psiDestinationFolder),
            get_name(psiNewlyCreated),
            hrMove,
        )

    @_err_to_hresult
    def PreCopyItem(
        self,
        dwFlags: int,
        psiItem: IShellItem,
        psiDestinationFolder: IShellItem,
        pszNewName: str,
    ):
        self.impl.pre_copy_item(
            TransferSourceFlags(dwFlags),
            get_name(psiItem),
            get_name(psiDestinationFolder),
            pszNewName,
        )

    @_err_to_hresult
    def PostCopyItem(
        self,
        dwFlags: int,
        psiItem: IShellItem,
        psiDestinationFolder: IShellItem,
        pszNewName: str,
        hrCopy: int,
        psiNewlyCreated: IShellItem | None,
    ):
        self.impl.post_copy_item(
            TransferSourceFlags(dwFlags),
            get_name(psiItem),
            get_name(psiDestinationFolder),
            get_name(psiNewlyCreated),
            hrCopy,
        )

    @_err_to_hresult
    def PreDeleteItem(self, dwFlags: int, psiItem: IShellItem):
        self.impl.pre_delete_item(TransferSourceFlags(dwFlags), get_name(psiItem))

    @_err_to_hresult
    def PostDeleteItem(
        self,
        dwFlags: int,
        psiItem: IShellItem,
        hrDelete: int,
        psiNewlyCreated: IShellItem | None,
    ):
        recycled = False if not psiNewlyCreated else True
        self.impl.post_delete_item(
            TransferSourceFlags(dwFlags), get_name(psiItem), hrDelete, recycled
        )

    @_err_to_hresult
    def PreNewItem(
        self, dwFlags: int, psiDestinationFolder: IShellItem, pszNewName: str
    ):
        self.impl.pre_new_item(
            TransferSourceFlags(dwFlags), get_name(psiDestinationFolder), pszNewName
        )

    @_err_to_hresult
    def PostNewItem(
        self,
        dwFlags: int,
        psiDestinationFolder: IShellItem,
        pszNewName: str,
        pszTemplateName: str,
        dwFileAttributes: int,
        hrNew: int,
        psiNewItem: IShellItem | None,
    ):
        self.impl.post_new_item(
            TransferSourceFlags(dwFlags),
            get_name(psiDestinationFolder),
            get_name(psiNewItem),
            FileAttributeFlags(dwFileAttributes),
            hrNew,
        )

    @_err_to_hresult
    def UpdateProgress(self, iWorkTotal: int, iWorkSoFar: int):
        self.impl.update_progress(iWorkTotal, iWorkSoFar)

    @_err_to_hresult
    def ResetTimer(self):
        self.impl.reset_timer()

    @_err_to_hresult
    def PauseTimer(self):
        self.impl.pause_timer()


class FileOperationProgressSink:
    """The base class for user-implemented IFileOperationProgressSink objects.

    Subclass and implement any methods for which you wish to perform additional tasks.
    In each method, to indicate an error, raise an exception (rather than returning a
    WIN32 error code).  For the post_* methods, the `new_name` paramter will be `None`
    if the operation failed, or the full path of the new object on a success.

    Note that the pre_* and post_* methods may be called multiple times per file,
    depending on the flags specified in the IFileOperation object. For example, if
    confirmation dialogs are enabled (default), and a name collision occurs, the post_*
    method for the operation can be called multiple times with a PENDING status. Once
    the user has selected a final action, you will know if it succeeded or not by
    checking `new_name` for None.

    The `result` parameter is a WIN32 HRESULT value indicating the result of the
    operation. Currently, you should only need to check this for post_delete_item, as
    there is no good way to check for success otherwise.
    """

    @cached_property
    def com_ptr(self):
        """Return a pywin32 compatible COM pointer for passing to other pywin32
        COM objects.
        """
        return com_ptr(_FileOperationProgressSink(self))

    def start_operations(self) -> None:
        """Called when the file operations are about to begin."""

    def finish_operations(self, result: int) -> None:
        """Called when all file operations are completed. `result` is a WIN32 HRESULT
        with the final result of the operation.
        """

    def pre_rename_item(
        self, transfer_flags: TransferSourceFlags, source: str, new_name: str
    ) -> None:
        """Called just prior to renaming an item."""

    def post_rename_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        new_name: str | None,
        result: int,
    ) -> None:
        """Called after an item has been renamed."""

    def pre_move_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        destination: str,
        new_name: str,
    ) -> None:
        """Called just prior to moving an item."""

    def post_move_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        destination: str,
        new_name: str | None,
        result: int,
    ) -> None:
        """Called after an item has been moved."""

    def pre_copy_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        destination: str,
        new_name: str,
    ) -> None:
        """Called just prior to copying an item."""

    def post_copy_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        destination: str,
        new_name: str | None,
        result: int,
    ) -> None:
        """Called after an item has been copied."""

    def pre_delete_item(self, transfer_flags: TransferSourceFlags, source: str) -> None:
        """Called just prior to deleting an item."""

    def post_delete_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        result: int,
        recycled: bool,
    ) -> None:
        """Called after an item has been deleted. `recycled` is True if the item was
        sent to the recycle bin.
        """

    def pre_new_item(
        self, transfer_flags: TransferSourceFlags, destination: str, new_name: str
    ) -> None:
        """Called just prior to creating a new item."""

    def post_new_item(
        self,
        transfer_flags: TransferSourceFlags,
        destination: str,
        new_name: str | None,
        file_attributes: FileAttributeFlags,
        result: int,
    ) -> None:
        """Called after a new item has been created."""

    def update_progress(self, work_total: int, work_so_far: int) -> None:
        """Called periodically to update the progress of the operation."""

    def reset_timer(self) -> None:
        """Called when the timer is reset."""

    def pause_timer(self) -> None:
        """Called when the timer is paused."""

    def resume_timer(self) -> None:
        """Called when the timer is resumed."""



__all__ = [
    'FileOperator',
]

from os import PathLike, fspath
from pathlib import Path
from typing import Iterable, TypeAlias

from .com import (
    FileOperation,
    FileOperationProgressSink,
    convert_exceptions,
    parse_filename,
)
from .errors import IFileOperationError, UnexpectedError
from .flags import FileOperationFlags, FileOperationResult, TransferSourceFlags

StrPath: TypeAlias = str | PathLike[str]


class ProgressSink(FileOperationProgressSink):
    """ProgressSink that logs successful operations."""

    def __init__(self):
        self.name_map = {}

    def finish_operations(self, result: int) -> None:
        self.result_code = result

    def post_copy_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        destination: str,
        new_name: str | None,
        result: int,
    ) -> None:
        if new_name:
            self.name_map[source] = new_name

    def post_delete_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        result: int,
        recycled: bool,
    ) -> None:
        if result == FileOperationResult.SUCCESS:
            if recycled:
                self.name_map[source] = 'RECYCLE_BIN'
            else:
                self.name_map[source] = 'DELETED'

    def post_move_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        destination: str,
        new_name: str | None,
        result: int,
    ) -> None:
        if new_name:
            self.name_map[source] = new_name

    def post_rename_item(
        self,
        transfer_flags: TransferSourceFlags,
        source: str,
        new_name: str | None,
        result: int,
    ) -> None:
        if new_name:
            self.name_map[source] = new_name


class FileOperator:
    """A context manager for performing file operations.  Not reentrant."""

    DEFAULT_FLAGS = None
    """Default file operation behavior. File operations behave just as if the user had
    performed them in Windows Explorer with no modifiers (eg: Shift, Ctrl, etc).
    """

    UNDO_FLAGS = (
        FileOperationFlags.ADDUNDORECORD
        | FileOperationFlags.ALLOW_UNDO
        | FileOperationFlags.RECYCLEONDELETE
    )
    """Standard flags to allow operations to be undone with Ctrl + Z."""

    SEMI_SILENT_FLAGS = (
        FileOperationFlags.WANTNUKEWARNING
        | FileOperationFlags.SILENT
        | FileOperationFlags.NO_CONFIRM_MKDIR
    )
    """Flags to suppress progress dialogs, but still show dialogs for name collisions
    and temporary errors (eg: due to a file open in another program), allowing the user
    to Try Again or Skip.
    """

    FULL_SILENT_FLAGS = (
        FileOperationFlags.SILENT
        | FileOperationFlags.NO_CONFIMATION
        | FileOperationFlags.NOERRORUI
        | FileOperationFlags.EARLYFAILURE
        | FileOperationFlags.NO_CONFIRM_MKDIR
        | FileOperationFlags.SHOWELEVATIONPROMPT
    )
    """Flags to suppress all dialogs (as if Yes to All was selected), except for a UAC
    prompt if necessary.  If errors occur, the operation will fail immediately.
    """

    @convert_exceptions
    def __init__(
        self,
        parent=None,
        flags: FileOperationFlags | None = DEFAULT_FLAGS,
        *,
        commit_on_exit: bool = False,
    ):
        """Create a FileOperator instance for managing file operations.  To be used with
        the `with` statement. Operations are scheduled first, then performed with the
        `commit` method, or optionally automatically when the with block is exited.

        The context manager is not reentrant.

        :param parent: A HANDLE to the window that should own any dialog boxes that are
            displayed. Also supports wx.Window instances.
        :param flags: The FileOperationFlags to use for the operations.
        :param commit_on_exit: If True (default: False), `commit` will be performed
            automatically when the with block is exited if no exceptions were raised.
        """
        self.ifo = FileOperation()
        self.sink = ProgressSink()
        self.sink_cookie = self.ifo.Advise(self.sink.com_ptr)
        if parent is not None:
            try:
                parent = parent.GetHandle()  # wx.Window
            except AttributeError:
                pass
        if parent is not None:
            self.ifo.SetOwnerWindow(parent)
        if flags is not None:
            self.ifo.SetOperationFlags(int(flags))
        self.commit_on_exit = commit_on_exit
        self.entered = False
        self._operations_queued = False

    def __enter__(self):
        if self.entered:
            raise IFileOperationError(f'{type(self).__name__} is not reentrant')
        else:
            self.entered = True
            return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.commit_on_exit and not exc_type:
            self.commit()
            self.ifo.Unadvise(self.sink_cookie)
        # Delete the IFileOperation instance, so reentrance will raise an error
        del self.ifo

    @convert_exceptions
    def move_file(
        self, source: StrPath, destination: StrPath, new_name: str | None = None
    ) -> None:
        """Schedule to move a file from `source` to `destination`.  Optionally rename
        the file in the process. NOTE: Windows performs a move to a different logical
        drive as a copy and delete.

        :param source: The file to be moved.
        :param destination: The destination *directory* to move to.
        :param new_name: The new name for the file if desired.
        """
        self.ifo.MoveItem(
            parse_filename(source), parse_filename(destination, True), new_name
        )
        self._operations_queued = True

    @convert_exceptions
    def move_files(self, sources: Iterable[StrPath], destination: StrPath) -> None:
        """Schedule to move the files `sources` to `destination`. NOTE: Windows performs
        a move to a different logical drive as a copy and delete.

        :param sources: And iterable of filenames to be moved.
        :param destination: The destination *directory* to move to.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.MoveItems(srcs, parse_filename(destination, True))
        self._operations_queued = True

    @convert_exceptions
    def rename_file(
        self, source: StrPath, new_name: str, allow_move: bool = True
    ) -> None:
        """Schedule to rename `source` to `new_name`.  If `allow_move` is True and the
        destination is in a different directory, this will perform a move instead.

        :param source: The file to be renamed.
        :param new_name: The new name of the file.
        :param allow_move: If True, then converts a rename into a move when new_name is
            in a different directory than the source.
        """
        if allow_move:
            # Check if new_name refers to a file in a different directory
            dest_path = Path(new_name)
            if len((dest_parts := dest_path.parts)) == 1:
                # No directory on the new_name, all good for a rename
                pass
            elif Path(source).parts[:-1] != dest_parts[:-1]:
                # new_name wants the file to be in a different directory, do a move
                # instead
                self.move_file(source, dest_path.parent, dest_path.name)
                return
            else:
                # new_name includes the directory, but it's the same directory, so a
                # rename is fine
                new_name = dest_path.name
        self.ifo.RenameItem(parse_filename(source), new_name)
        self._operations_queued = True

    @convert_exceptions
    def rename_files(self, sources: Iterable[StrPath], new_name: StrPath) -> None:
        """Schedule to rename the files `sources` to `new_name`.  All files will be
        renamed to the new name.

        :param sources: An iterable of filenames to be renamed.
        :param new_name: The new name of the files.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.RenameItems(srcs, fspath(new_name))
        self._operations_queued = True

    @convert_exceptions
    def copy_file(
        self, source: StrPath, destination: StrPath, new_name: str | None = None
    ) -> None:
        """Schedule to copy a file from `source` to `destination`. Optionally rename the
        file in the process.

        :param source: The file to be copied.
        :param destination: The destination *directory* to copy to.
        :param new_name: Optional new name of the file.
        """
        self.ifo.CopyItem(
            parse_filename(source), parse_filename(destination, True), new_name
        )
        self._operations_queued = True

    @convert_exceptions
    def copy_files(self, sources: Iterable[StrPath], destination: StrPath) -> None:
        """Schedule to copy the files `sources` to `destination`.

        :param sources: An iterable of filenames to be copied.
        :param destination: The destination *directory* to copy to.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.CopyItems(srcs, parse_filename(destination, True))
        self._operations_queued = True

    @convert_exceptions
    def delete_file(self, source: StrPath) -> None:
        """Schedule to delete a file.

        :param source: The file to be deleted.
        """
        self.ifo.DeleteItem(parse_filename(source))
        self._operations_queued = True

    @convert_exceptions
    def delete_files(self, sources: Iterable[StrPath]) -> None:
        """Schedule to delete the files `sources`.

        :param sources: An iterable of filenames to be deleted.
        """
        srcs = [parse_filename(s) for s in sources]
        self.ifo.DeleteItems(srcs)
        self._operations_queued = True

    @convert_exceptions
    def _perform_operations(self):
        self.return_code = self.ifo.PerformOperations()

    def commit(self):
        """Perform all scheduled file operations."""
        try:
            self._perform_operations()
        except UnexpectedError:
            if not self._operations_queued:
                # Actually expected when no operations have been queued
                self.return_code = 0
                self.aborted = False
                self.results = {}
            else:
                raise
        else:
            self.aborted = self.ifo.GetAnyOperationsAborted()
            self.results = self.sink.name_map
            if self.return_code is None:
                self.return_code = self.sink.result_code
